/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/********************************************************************************/
/*      ________________________________________________________________        */
/*      | H Y T E C   8 4 1 4  1 6 B i t   A D C  A s y n  D r i v e r |        */
/*           -----------------------------------------------------------        */
/*                                                                              */
/*  Source file name      :-      drvHy8414Asyn.cpp                             */
/*                                                                              */
/*  Initial creation date :-      15-Apr-2011                                   */
/*                                                                              */
/*  Original Developers   :-      Jim Chen						                */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Description :- This code is for Hytec 8414 ADC EPICS support. It is based   */
/*  on Hy8417Asyn driver code. It is osi complaint.		                    	*/
/*                                                                              */
/*  (C)2011 Hytec Electronics Ltd.                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/* Modification record :-                                                       */
/*                                                                              */
/*  Date modified |  Person   | Version | Description of modification           */
/* ---------------|-----------|---------|-------------------------------------- */
/*    15/04/11    |  JC		  | 2.3     | Initial creation with new asyn model  */    
/*    05/12/11    |  JC		  | 2.4     | Changed to only support 8 channels    */    
/*                |           |         | for the new design.                   */    
/*    16/12/11    |  JC		  | 2.5     | Fixed the parameter display wrong	    */    
/*                |           |         | at the start                          */    
/*    19/01/12    |  JC		  | 2.6     | Fixed +/-5V setting not being set	    */    
/*                |           |         | properly		                        */    
/*    24/01/12    |  JC		  | 2.7     | Changed returning the parameters 	    */    
/*                |           |         | by calling the base class readInt32   */    
/*    13/02/14    |  JC		  | 3.0     | Converted to asyn model 3				*/    
/*                |           |         | Now use 2M memory 					*/    
/*    06/10/14    |  JC		  | 3.1     | Modified to support 64bit Linux based	*/    
/*                |  M.Rivers |         | on Mark Rivers' driverTable model     */    
/*                |           |         | Note: need V2.6 of drvHyLinuxDriver.c */    
/*                |           |         | in ipac module to work with 64bit     */    
/*                |           |         | Linux		                            */    
/*                |           |         |    		                            */    
/********************************************************************************/

#define DRIVER_VERSION	301

/******************/
/* Hytec Module includes */
/******************/
#include "drvHy8414Asyn.h"

static Hy8414Asyn *driverTable[MAX_CARDS];
static int numCards;

/********************************************************************/
/*														           */
/* Public Sector												  */
/*													             */
/****************************************************************/
/* Constructor:
 *
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *    vectorNum = interrupt vector (0 = find one for me),
 *      samples = number of samples (not applicable in gated modes),
 *		average = number of readings to SUM for average
 *       offset = Trigger mode: Offset from trigger point to first ADC reading to return
 *				  Gated mode:   Offset from the start (or end if negative) of the gated period to the ADC readings to return
 *				  Continuous:   Not applicable
 *         mode = 0 = continuous, 1 = triggered, 3 = gated,
 *    clockRate = 0  = 1Hz,    1 = 2Hz,    2 = 5Hz,    3 = 10Hz,
 *				  4  = 20Hz,   5 = 50Hz,   6 = 100Hz,  7 = 200Hz,
 *				  8  = 500Hz,  9 = 1kHz,  10 = 2kHz,  11 = 5kHz,
 *				  12 = 10khz, 13 = 20Khz, 14 = 50kHz, 15 = 100kHz.
 *     extClock = 0 = internal, 1 = extrenal,
 *		fastAdc = flag to say if the ADC is running at fast speed for normal use. =1 fast, =0 slow for MCA & EPID records 
 *		  range = 0 : +/-10V, 1 : +/-5V
 *   ChannelBit = 16 : 16 bit per channel; 24 : 24 bit per channel
 */
    
Hy8414Asyn::Hy8414Asyn(char *portName, int carrierNum, int ipSlotNum, int vectorNum, int mode)    
		: asynPortDriver(portName, MAX_ADC_CHANNELS, NUM_PARAMS, 
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask | asynDrvUserMask,
        asynInt32Mask | asynInt32ArrayMask | asynFloat64Mask | asynFloat64ArrayMask, 
        ASYN_MULTIDEVICE, 1, /* ASYN_CANBLOCK=0, ASYN_MULTIDEVICE=1, autoConnect=1 */
        0, 0)  /* Default priority and stack size */

{
    const char *hytecstr = "(Hytec Electronics Ltd., Reading, UK)";
    int strok;
  	char lstr[7];
  	USHORT idmem;
    int ishytec;
    int ismodel;
	int ver;

	this->lastPointer = 0;

	/* sweep data channels definition */
    createParam(hy8414SweepDataString, 		asynParamInt32, 	&hy8414SweepData_);
    createParam(hy8414FloatSweepDataString, asynParamFloat64, 	&hy8414FloatSweepData_);

	/* normal data channels definition */
    createParam(hy8414DataString, 			asynParamInt32, 	&hy8414Data_);
    createParam(hy8414FloatDataString, 		asynParamFloat64, 	&hy8414FloatData_);

	/* controls definition */
    createParam(hy8414ScanPeriodString, 	asynParamFloat64, 	&hy8414ScanPeriod_);
    createParam(hy8414SetSampleString, 		asynParamInt32, 	&hy8414SetSample_);
    createParam(hy8414SetAverageString, 	asynParamInt32, 	&hy8414SetAverage_);
    createParam(hy8414SetModeString, 		asynParamInt32, 	&hy8414SetMode_);
    createParam(hy8414SetOffsetString, 		asynParamInt32, 	&hy8414SetOffset_);
    createParam(hy8414EnableString, 		asynParamInt32, 	&hy8414Enable_);
    createParam(hy8414SoftTriggerString, 	asynParamInt32, 	&hy8414SoftTrigger_);
    createParam(hy8414AutoReenableString, 	asynParamInt32, 	&hy8414AutoReenable_);
    createParam(hy8414ClearBufferString, 	asynParamInt32, 	&hy8414ClearBuffer_);
    createParam(hy8414SetClockRateString, 	asynParamInt32, 	&hy8414SetClockRate_);
    createParam(hy8414SetExtClockString, 	asynParamInt32, 	&hy8414SetExtClock_);
    createParam(hy8414CsrString, 			asynParamInt32, 	&hy8414Csr_);
    createParam(hy8414ExtCsrString, 		asynParamInt32, 	&hy8414ExtCsr_);

    createParam(hy8414GainString, 			asynParamInt32, 	&hy8414Gain_);
    createParam(hy8414RangeString, 			asynParamInt32, 	&hy8414Range_);
    createParam(hy8414FastADCString, 		asynParamInt32, 	&hy8414FastADC_);

    createParam(hy8414AverageOverflowString,asynParamInt32, 	&hy8414AverageOverflow_);
    createParam(hy8414OverflowString, 		asynParamInt32, 	&hy8414Overflow_);
    createParam(hy8414BufferCountString, 	asynParamInt32, 	&hy8414BufferCount_);
    createParam(hy8414TriggerStatusString,	asynParamInt32, 	&hy8414TriggerStatus_);
    createParam(hy8414FWVersionString, 		asynParamInt32, 	&hy8414FWVersion_);
    createParam(hy8414DriverVersionString,	asynParamInt32, 	&hy8414DriverVersion_);
    createParam(hy8414SupportString, 		asynParamInt32, 	&hy8414Support_);
	
	/* Save the parameters in the driver structure    */
    this->carrierNum		 = carrierNum;
    this->ipSlotNum			 = ipSlotNum;
    this->vectorNum          = vectorNum;
	
	setIntegerParam(hy8414SetMode_, mode);
	this->mode = mode;

	/* Map Register address spaces  */
	this->adc = (IO_MEMORY *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrIO);
    if (this->adc == NULL)
    {  
        /* can't register card register address space */
    	printf("configHy8414: Unable to register IO space\n");
        return;
    }

    /* map the memory address spaces  */
    this->convMemoryBase = (USHORT *) ipmBaseAddr(carrierNum, ipSlotNum, ipac_addrMem);
    if (this->convMemoryBase == NULL)
    {  
        /* can't register card Memory address space */
     	printf("configHy8414: Unable to register memory space\n");
        return;
    }
    
    /* Compare the HY8414 prom header, manufacturer's ID and card model
     * number with the expected values. */
  	idmem =  *((USHORT*)this->adc->prom.vita4);
  	if((idmem == 0x5649) || (idmem == 0x4956));
  	{
  		lstr[0] = 0x56; lstr[1] = 0x49;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 1);
  	if((idmem == 0x5441) || (idmem == 0x4154));
  	{	
  		lstr[2] = 0x54; lstr[3] = 0x41;
  	}
  	idmem =  *((USHORT*)this->adc->prom.vita4 + 2);
  	if((idmem == 0x3420) || (idmem == 0x2034));
  	{	
  		lstr[4] = 0x20; lstr[5] = 0x34;
  	}
  	lstr[6] = 0;

    strok = memcmp(lstr, VITA4, sizeof(this->adc->prom.vita4)) == 0;
    ishytec = this->adc->prom.hytecIdHigh == HYTECID_HI  &&
                  this->adc->prom.hytecIdLow == HYTECID_LO;
    ismodel = this->adc->prom.model == PROM_MODEL;

	/* check if we have been asked to be more chatty */
    /* start with some basic information */
    printf("PROM: Contents at offset - %#08x\n", (int)&this->adc->prom);
    printf("PROM: Prom header string - '%6s'\n", lstr);
    printf("PROM: Manufacturer's ID  - 0x%04x%04x %s\n",
        (int)this->adc->prom.hytecIdHigh, (int)this->adc->prom.hytecIdLow, 
        ishytec ? hytecstr : "(Unknown)");
    printf("\n");
    /* output a bit more about this specific model and card */
    printf("PROM: Model #/revision   - %#04hx rev. %#04hx\n",
        (int)this->adc->prom.model, (int)this->adc->prom.revision);

	ver = (int)this->adc->prom.revision;
	ver = ver/4096*1000 + (ver%4096)/256*100 + ((ver%4096)%256)/16*10 + ((ver%4096)%256)%16;
	setIntegerParam(hy8414FWVersion_, ver);

    printf("PROM: Serial number      - %hu\n", (int)this->adc->prom.serialNumber);
    printf("PROM: Driver IDs         - %#04hx, %#04hx\n",
        (int)this->adc->prom.driverIdHigh, (int)this->adc->prom.driverIdLow);
    printf("PROM: Device flags       - %#04hx\n", (int)this->adc->prom.flags);
    printf("PROM: Bytes used         - %#04hx (%hd)\n",
        (int)this->adc->prom.byteCount, (int)this->adc->prom.byteCount);
    printf("\n");

    /* check if everything passed okay */
    if (!strok)
        printf("PROM: INVALID PROM HEADER; EXPECTED '%s'\n", VITA4);
    if (!ishytec)
        printf("PROM: UNSUPPORTED MANUFACTURER ID; EXPECTED 0x%04x%04x, %s\n",
            HYTECID_HI, HYTECID_LO, hytecstr);
    if (!ismodel)
	{
        printf("PROM: UNSUPPORTED MODEL NUMBER; EXPECTED %#04hx\n",
            PROM_MODEL);
	}

    /* clear asyn message sent counters */
    this->messagesSent = 0;
    this->messagesFailed = 0;

    /* set up interrupt, work queue and mutex stuff */
    this->adc->reg.vector = vectorNum;            /* writes vector register offset 0x08 */
    driverTable[numCards] = this;
    numCards++;
	if (ipmIntConnect(this->carrierNum, this->ipSlotNum, this->vectorNum, INTSRC, numCards-1) < 0) {
      printf("configHy8414: failure connecting to interrupt\n");
      return;
    }

    ipmIrqCmd(this->carrierNum, this->ipSlotNum, 0, ipac_irqEnable);

    this->intMsgQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8414intQueuedTask",
                           epicsThreadPriorityLow,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intQueuedTaskC,
                           (void *)this) == NULL)
	{	
      	printf("configHy8414: failure creating message queue.\n");
      	return;
	}
    
    this->intSweepQId = epicsMessageQueueCreate(MAX_MESSAGES, 
                                              (MAX_ADC_CHANNELS + 2)*sizeof(int));
    if (epicsThreadCreate("Hy8414intSweepQTask",
                           epicsThreadPriorityHigh,
                           epicsThreadGetStackSize(epicsThreadStackMedium),
                           (EPICSTHREADFUNC)intSweepQueuedTaskC,
                           (void *)this) == NULL)
	{
       	printf("Hy8414intSweepQueuedTask epicsThreadCreate failure\n");
      	return;
	}

	this->rebooting = 0;

   	epicsAtExit(rebootCallbackC, this);

}

// Extension Init
int Hy8414Asyn::Hy8414ExtInit(int sample, int average, int offset, int clockRate, 
		int extClock, int fastADC, int gain, int range)
{
	
	setIntegerParam(hy8414SetSample_, sample);
	this->samples = sample;
	setIntegerParam(hy8414SetAverage_, average);
	setIntegerParam(hy8414SetOffset_, offset);
	setIntegerParam(hy8414SetClockRate_, clockRate);
	setIntegerParam(hy8414SetExtClock_, extClock);
	setIntegerParam(hy8414TriggerStatus_, WAITING_TRIGGER);
	setIntegerParam(hy8414SoftTrigger_, 0);
	setIntegerParam(hy8414Enable_, 0);
	this->enable = 0;
	setIntegerParam(hy8414AutoReenable_, 0);
	setIntegerParam(hy8414BufferCount_, 0);
	setIntegerParam(hy8414FastADC_, fastADC);
	this->fastadc = fastADC;
	setIntegerParam(hy8414Gain_, gain);
	setIntegerParam(hy8414Range_, range);
	setIntegerParam(hy8414DriverVersion_, DRIVER_VERSION);

	this->banksize = BANKSIZE;

	/*******************************************************************/
	/************* Working out real stuffs for ADC *********************/
	/*******************************************************************/

	setupADC( INIT );

	// tell listeners the changes
	callParamCallbacks();

	return(asynSuccess);
}


/*********************************************************/
/*                                                      */
/* Asyn interface Function implementations             */
/*                                                    */
/*****************************************************/

/* asynCommon methods */
void Hy8414Asyn::report(FILE *fp, int details)
{
	int mode, range, extclock, clockrate, sample, average, offset, fwver, swver, fastadc, gain;

    fprintf(fp, "Port: %s, carrier slot %d slot %d, base address=%p\n", 
            this->portName, this->carrierNum, this->ipSlotNum, (void *)this->adc);
    if (details >= 1) 
    {
		getIntegerParam(hy8414SetSample_, &sample);
		getIntegerParam(hy8414SetAverage_, &average);
		getIntegerParam(hy8414SetMode_, &mode);
		getIntegerParam(hy8414SetOffset_, &offset);
		getIntegerParam(hy8414SetClockRate_, &clockrate);
		getIntegerParam(hy8414SetExtClock_, &extclock);
		getIntegerParam(hy8414FWVersion_, &fwver);
		getIntegerParam(hy8414FastADC_, &fastadc);
		getIntegerParam(hy8414Gain_, &gain);
		getIntegerParam(hy8414Range_, &range);
		getIntegerParam(hy8414DriverVersion_, &swver);
		
        fprintf(fp, "  INT Vector=%x\n", this->vectorNum);
        fprintf(fp, "  Sample=%d\n", sample);
        fprintf(fp, "  Average=%d\n", average);
        fprintf(fp, "  Mode=%s\n", mode==0?"Normal":(mode==1?"Trigger":"Gated"));
        fprintf(fp, "  Offset=%d\n", offset);
        fprintf(fp, "  Clock Source=%s\n", extclock==0?"Internal":"External");
        fprintf(fp, "  Clock Rate=%d\n", clockrate);
        fprintf(fp, "  Gain=%d\n", gain);
        fprintf(fp, "  Range=%s\n", range==0?"+/-10V":"+/-5V");
        fprintf(fp, "  Fast ADC=%s\n", fastadc==1?"Yes":"No");
        fprintf(fp, "  Firmware Version=%d\n", fwver);
        fprintf(fp, "  Driver Version=%d\n", swver);
        fprintf(fp, "  messages sent OK=%d; send failed (queue full)=%d\n",
                this->messagesSent, this->messagesFailed);       
    }
    asynPortDriver::report(fp, details);
}


/*****         asynInt32 methods       ******/
asynStatus Hy8414Asyn::writeInt32(asynUser *pasynUser, epicsInt32 value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;

    this->getAddress(pasynUser, &channel);
	
/*	printf("\n*** hy8414_int32Write *** command = %d (%s), channel = %d val = %i\n\n",
				command, hy8414_commands[command].commandString, channel, value);
*/	
    status = setIntegerParam(function, value);

	/* decide what to do depending on which command we received */
	if (function == hy8414SetMode_)
	{
		this->mode = value;
        setupADC( RUNTIME );
	}
	else if (function == hy8414Enable_)
		setEnableDisable( value );
	else if (function == hy8414SetAverage_)
        checkAverageOverflow( RUNTIME );
	else if (function == hy8414SetSample_)
	{
		this->samples = value;
        setSamples( RUNTIME );	
	}
	else if (function == hy8414SetOffset_)
        setOffset( RUNTIME );
	else if (function == hy8414SoftTrigger_)
        softwareTrigger();
	else if (function == hy8414SetClockRate_)
        setClockRate( RUNTIME );
	else if (function == hy8414SetExtClock_)
        setExtClock( RUNTIME );
	else if (function == hy8414Range_)
        setRange( value );
	else if (function == hy8414Gain_)
        setGain( value );

	// tell listeners the changes
  	callParamCallbacks();

	return( status );
}

asynStatus Hy8414Asyn::readInt32(asynUser *pasynUser, epicsInt32 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel, localValue=0;
    int i, pointer, numToAverage = 0, startPoint = 0, contAverage = 0, half_mem = 0;
    double dValue = 0.0;
	int function = pasynUser->reason;
	int mode, average, samples, offset, triggerState;
    if (this->rebooting) epicsThreadSuspendSelf();
   	
	getIntegerParam(hy8414SetSample_, &samples);
	getIntegerParam(hy8414SetAverage_, &average);
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);
	getIntegerParam(hy8414TriggerStatus_, &triggerState);

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8414Data_)
	{
    	pointer = this->adc->reg.pointer;

        /* Note: average is already checked and corrected during the initialisation 
         * by checkAverageOverflow routine for continuous and trigger mode */
        switch(mode)
        {
        	case hy8414Continuous:
            	if (average > samples)
                {
    				status = setIntegerParam(hy8414AverageOverflow_, 1);					/* averaging overflow, this is an abnormal averaging */
                    contAverage = samples;
                }else
                {
                    status = setIntegerParam(hy8414AverageOverflow_, 0);        			/* averaging normal */
                    contAverage = average;
                }
                if (pointer >= contAverage) 
                    startPoint = pointer - contAverage;
                else
                    startPoint = this->banksize * 2 - (contAverage - pointer);
                numToAverage = contAverage;
                break;
            case hy8414Triggered:
                if (pointer == 0) pointer = this->lastPointer; 		/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (pointer <= offset)                                    					/* current pointer less than offset, take whatever is collected. ignore offset */
                {
                    if (pointer < average)                                					/* current pointer also less than average */
                    {
                        startPoint = 0;                                         			/* use all the data available */
                        numToAverage = pointer;
                    }else                                                       			/* cuurent pointer greater than average */
                    {
                        startPoint = pointer  - average;          					/* use most recent data */
                        numToAverage = average;
                    }   
                    status = setIntegerParam(hy8414AverageOverflow_, 1);        			/* averaging overflow, this is not normal averaging */
                } else                                                          			/* current pointer is greater than offset */
                {
                    if ((pointer - offset) >= average)              						/* if pointer minus offset is greater than average, take most recent data to average */
                    { 
                        startPoint = pointer - average;
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 0);    			/* This is the only normal averaging. */
                    }else                                                       			/* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                    {
                        startPoint = offset;
                        numToAverage = pointer - offset;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);    			/* averaging overflow, this is an abnormal averaging */
                    }
                }
                break;
            case hy8414Gated:
                if (pointer == 0) pointer = this->lastPointer;		/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
            	if (offset >= 0)                                           					/* offset is possitive */
                {
                    if (pointer < offset)                                 					/* current pointer less than offset, take whatever is collected. ignore offset */
                    {
                        if (pointer < average)                            					/* current pointer also less than average */
                        {
                            startPoint = 0;                                     			/* use all the data available */
                            numToAverage = pointer;
                        }else                                                   			/* cuurent pointer greater than average */
                        {
                            startPoint = pointer - average;     					/* use most recent data */
                            numToAverage = average;
                        }   
                        status = setIntegerParam(hy8414AverageOverflow_, 1);    			/* averaging overflow, this is not normal averaging */
                    } else                                                      			/* current pointer is greater than offset */
                    {
                        if ((pointer - offset) >= average)          						/* if pointer minus offset is greater than average, take most recent data to average */
                        { 
                            startPoint = pointer - average;
                            numToAverage = average;
                            status = setIntegerParam(hy8414AverageOverflow_, 0);			/* This is the only normal averaging */
                        }else                                                   			/* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                        {
                            startPoint = offset;
                            numToAverage = pointer - offset;	
                            status = setIntegerParam(hy8414AverageOverflow_, 1);			/* averaging overflow, this is an abnormal averaging */
                        }
                    }
                }else                                                           			/* offset is negative */
                {
                    if (pointer >= (average-offset))										//offset is minus number
                    {
                        startPoint = pointer - average + offset; 					/* take whatever available */
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 0);				/* averaging normal */
                    }else if(pointer >= average)
                    {
                        startPoint = 0; 													/* take whatever available */
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);				/* averaging not normal */                    	
                    }else
                    {
                        startPoint = 0; 													/* take whatever available */
                        numToAverage = pointer;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);				/* averaging not normal */
                    }
                }
                break;
            default:
                status = asynError;
                break;
        }
		
/*if(channel == 0)
printf("c=%d csr=0x%x extcsr=0x%x pnt=%d stpnt=%d notoaverage=%d  \n",
		channel, this->adc->reg.status, this->adc->reg.extStatus, pointer, 
		startPoint, numToAverage);                              
for debug */
		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

		for(i=0; i<numToAverage; i++)
        {
            dValue += (double) ((USHORT) ((this->convMemoryBase + half_mem + channel * this->banksize)[startPoint]));
            startPoint++;
        	
            if(startPoint >= this->banksize)
            {
            	startPoint = 0;                          									/* if pointer reaches the end of the either the full memory of half full, reset it */
				half_mem = half_mem == 0 ? HALF_MEM : 0;
            }
        }
        if (numToAverage != 0)
            localValue = (epicsInt32) (dValue / numToAverage);                  			/* check deviding 0 for safety */
        else
            localValue = (epicsInt32) dValue; 
        *value = localValue;
/*
if(channel == 0)
printf("csr=%x ecsr=%x lv=0x%x-dv=%f\n", this->adc->reg.status, this->adc->reg.extStatus, localValue, dValue);
*/
	}
	else if(function == hy8414Support_)
        *value = DRIVER_SUPPORT;                                                /* driver support */
	else if(function == hy8414Csr_)
        *value = this->adc->reg.status;															/* csr */
	else if(function == hy8414ExtCsr_)
        *value = this->adc->reg.extStatus;															/* csr */
	else if(function == hy8414TriggerStatus_)
	{
        if (mode == hy8414Continuous) 
            *value = asynError;                                                 /* continuous mode doesn't support this */
        else if (mode == hy8414Triggered)
            localValue = this->adc->reg.extStatus & EXTCSR_TEN;
        else if (mode == hy8414Gated)
            localValue =  (~this->adc->reg.status & CSR_MIS);

        if (localValue > 0)
		{                                                     /* great than 0 means triggered */
            *value = TRIGGERRED_COLLECTING;
			setIntegerParam(hy8414TriggerStatus_, TRIGGERRED_COLLECTING);
		}
        else if (localValue == 0)                                               /* return whatever it is either waiting for trigger or finished or triigered by software */
            *value = triggerState;
    }else
	{
		asynPortDriver::readInt32(pasynUser, value);
	}
	
  	callParamCallbacks();
	return( status );
}

asynStatus Hy8414Asyn::getBounds(asynUser *pasynUser, epicsInt32 *low, epicsInt32 *high)
{
	*low = 0;
	*high = this->banksize*2;
	/* not usually implemented... See asyn documentation for details if needed. */
	return( asynSuccess );
}

/* asynInt32Array methods */
asynStatus Hy8414Asyn::readInt32Array(asynUser *pasynUser, epicsInt32 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, startPoint = 0, pointer, numToReturn = 0, half_mem = 0;
	int function = pasynUser->reason;
	int mode, offset;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);

	this->getAddress(pasynUser, &channel);
	
    if(function == hy8414Data_)
	{
        pointer = this->adc->reg.pointer;
        switch(mode)
        {
            case hy8414Continuous:
                if ((size_t) pointer >= nelements) 
                    startPoint = pointer - nelements;
                else
                    startPoint = this->banksize * 2 - (nelements - pointer);
                numToReturn = nelements;
                break;
            case hy8414Triggered:
                if (pointer == 0) pointer = this->lastPointer;  	/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (pointer <= offset) 
                {
                    startPoint = 0;
                    numToReturn = pointer;
                } else
                {
                    startPoint = offset;
                    numToReturn = pointer - offset;
                }
                break;
            case hy8414Gated:
                if (pointer == 0) pointer = this->lastPointer;  /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (offset >= 0)
                {
                    if (pointer <= offset) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = offset;
                        numToReturn = pointer - offset;
                    }
                }
                else
                {
                    if (pointer < (-offset)) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = pointer + offset;
                        numToReturn = (-offset);
                    }
                }
                break;
            default:
                status = asynError;
                break;
        }
		
		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

/*for debug*/
//printf("chl=%d csr=%x ecsr=%x p=%d sp=%d hm=%x\n", channel, this->adc->reg.status, this->adc->reg.extStatus, pointer, startPoint, half_mem);
		
		for (i = 0; i<(int)nelements; i++)
        {
            if (i < numToReturn)
            {
                *(data + i) = (epicsInt32) (this->convMemoryBase + half_mem + channel * this->banksize)[startPoint];
               startPoint++;
            }
            else
                *(data + i) = 0;
            
            if(startPoint >= this->banksize)
            {
            	startPoint = 0;                          		/* if pointer reaches the bottom of the memory, reset it */
				half_mem = half_mem == 0 ? HALF_MEM : 0;
            }
        }
        if (nelements > (size_t)numToReturn)                                            /* if number to return less than number of elements, NORD = number of return */
            *nactual = numToReturn;
        else
            *nactual = nelements;                                               		/* if number to return is greater or equal number of elements, NORD = number of elements */
    }
    else if(function == hy8414SweepData_)
	{    	
    	pointer = this->adc->reg.pointer;
    	if ((mode == hy8414Triggered) || (mode == hy8414Gated))
        {
            if (pointer == 0) pointer = this->lastPointer;   /* for trigger/gated mode, use pointer stored by the interrupt only if adc address pointer is reset (=0) */
        }else
        {
        	if(pointer == 0)
        		pointer = this->banksize * 2 - 1;
        	else
        		pointer--;
        }

		half_mem = 0;
        if(pointer >= this->banksize)
        {
        	pointer = pointer - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

		for(i=0; i<(int)nelements; i++)
        {
       		*(data + i) = (epicsInt32) (this->convMemoryBase + half_mem + i * this->banksize)[pointer];
        }
        
        *nactual = nelements;
	}

	status = asynSuccess;
	return( status );
}

asynStatus Hy8414Asyn::readFloat64(asynUser *pasynUser, epicsFloat64 *value)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
	int function = pasynUser->reason;
    int i, pointer, numToAverage = 0, startPoint = 0, contAverage = 0, half_mem = 0;
    double dValue = 0.0, power = 0.0;
	int mode, average, samples, offset, period = 0, rate;

    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetSample_, &samples);
	getIntegerParam(hy8414SetAverage_, &average);
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);

	this->getAddress(pasynUser, &channel);
	
    if((function == hy8414Data_) || (function == hy8414FloatData_))
	{
    	pointer = this->adc->reg.pointer;
    	
        /* Note: average is already checked and corrected during the initialisation 
         * by checkAverageOverflow routine for continuous and trigger mode */
        switch(mode)
        {
            case hy8414Continuous:
                if (average > samples)
                {
                    status = setIntegerParam(hy8414AverageOverflow_, 1);		/* averaging overflow, this is an abnormal averaging */
                    contAverage = samples;
                }else
                {
                    status = setIntegerParam(hy8414AverageOverflow_, 0);        /* averaging normal */
                    contAverage = average;
                }
                if (pointer >= contAverage) 
                    startPoint = pointer - contAverage;
                else
                    startPoint = this->banksize * 2 - (contAverage - pointer);
                numToAverage = contAverage;
                break;
            case hy8414Triggered:
                if (pointer == 0) pointer = this->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (pointer < offset)                                     		/* current pointer less than offset, take whatever is collected. ignore offset */
                {
                    if (pointer < average)                                		/* current pointer also less than average */
                    {
                        startPoint = 0;                                         /* use all the data available */
                        numToAverage = pointer;
                    }else                                                       /* cuurent pointer greater than average */
                    {
                        startPoint = pointer - average;          		/* use most recent data */
                        numToAverage = average;
                    }   
                    status = setIntegerParam(hy8414AverageOverflow_, 1);       	/* averaging overflow, this is not normal averaging */
                } else                                                          /* current pointer is greater than offset */
                {
                    if ((pointer - offset) >= average)              			/* if pointer minus offset is greater than average, take most recent data to average */
                    { 
                        startPoint = pointer - average;
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 0);	/* This is the only normal averaging */
                    }else                                                       /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                    {
                        startPoint = offset;
                        numToAverage = pointer - offset;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);    /* averaging overflow, this is an abnormal averaging */
                    }
                }
                break;
            case hy8414Gated:
                if (pointer == 0) pointer = this->lastPointer;   /* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (offset >= 0)                                           		/* offset is possitive */
                {
                    if (pointer < offset)                                 		/* current pointer less than offset, take whatever is collected. ignore offset */
                    {
                        if (pointer < average)                            		/* current pointer also less than average */
                        {
                            startPoint = 0;                                     /* use all the data available */
                            numToAverage = pointer;
                        }else                                                   /* cuurent pointer greater than average */
                        {
                            startPoint = pointer - average;      		/* use most recent data */
                            numToAverage = average;
                        }   
                        status = setIntegerParam(hy8414AverageOverflow_, 1);    /* averaging overflow, this is not normal averaging */
                    } else                                                      /* current pointer is greater than offset */
                    {
                        if ((pointer - offset) >= average)          			/* if pointer minus offset is greater than average, take most recent data to average */
                        { 
                            startPoint = pointer - average;
                            numToAverage = average;
                            status = setIntegerParam(hy8414AverageOverflow_, 0);/* This is the only normal averaging */
                        }else                                                   /* if pointer minus offset is less than average, take the values starting from the offset to the end since this period is the user's interest */
                        {
                            startPoint = offset;
                            numToAverage = pointer - offset;
                            status = setIntegerParam(hy8414AverageOverflow_, 1);/* averaging overflow, this is an abnormal averaging */
                        }
                    }
                }else                                                           /* offset is negative */
                {
                    if (pointer >= (average-offset))										//offset is minus number
                    {
                        startPoint = pointer - average + offset; 					/* take whatever available */
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 0);				/* averaging normal */
                    }else if(pointer >= average)
                    {
                        startPoint = 0; 													/* take whatever available */
                        numToAverage = average;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);				/* averaging not normal */                    	
                    }else
                    {
                        startPoint = 0; 													/* take whatever available */
                        numToAverage = pointer;
                        status = setIntegerParam(hy8414AverageOverflow_, 1);				/* averaging not normal */
                    }
                }
                break;
            default:
                status = asynError;
                break;
        }

		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

		for(i=0; i<numToAverage; i++)
        {
/*        		
if(channel == 0) 
printf("i=%d csr=%x extcsr=%x p=%d sp=%d mem=%x h=%x l=%x regh=%x regl=%x\n",
     i, this->adc->reg.status, this->adc->reg.extStatus, pointer*words, startPoint, half_mem, highWord, lowWord, this->adc->reg.data[1], this->adc->reg.data[0]);
*/        		 
        	dValue += toVoltage((double) (this->convMemoryBase + half_mem + channel * this->banksize)[startPoint]);
        	startPoint++;        		
            if(startPoint >= this->banksize)
            {
            	startPoint = 0;                          		/* if pointer reaches the bottom of the memory, reset it */
				half_mem = half_mem == 0 ? HALF_MEM : 0;
            }
        }

        if (numToAverage != 0)                                                  /* safety check */
            *value = dValue / numToAverage;
        else
            *value = dValue;
    }
    else if((function == hy8414SweepData_) || (function == hy8414FloatSweepData_))                                       /* for FastSweep record, the most recent acquired data is used */
	{
    	pointer = this->adc->reg.pointer;
    	if ((mode == hy8414Triggered) || (mode == hy8414Gated))
        {
            if (pointer == 0) pointer = this->lastPointer;   /* for trigger/gated mode, use pointer stored by the interrupt only if adc address pointer is reset (=0) */
        }else
        {
        	if(pointer == 0)
        		pointer = this->banksize * 2 - 1;
        	else
        		pointer--;
        }

		half_mem = 0;
        if(pointer >= this->banksize)
        {
        	pointer = pointer - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

        dValue = (double) (this->convMemoryBase + half_mem + channel * this->banksize)[pointer];
    	
        *value = dValue;
	}
	// sort out scan_period
	else if (function == hy8414ScanPeriod_)
	{
		getIntegerParam(hy8414SetClockRate_, &rate);
        period = (int) rate % 3;
        if((period == 0) || (period == 2))
            period = period * 2 + 1;
        else
            period = 2;
        power = pow((double) 10, (double) ((int) (rate / 3)));
        dValue = (double) 1 / (power * period);                             /* this convert the frequency to second */
        *value = dValue;
        setDoubleParam(hy8414ScanPeriod_, dValue);
	}
	
  	callParamCallbacks();
	return( status );
}

/* asynFloat64Array methods */
asynStatus Hy8414Asyn::readFloat64Array(asynUser *pasynUser, epicsFloat64 *data, size_t nelements, size_t *nactual)
{
	asynStatus status = asynSuccess;
	epicsInt32 channel;
    int i, pointer, startPoint = 0, numToReturn = 0, half_mem = 0;
	int function = pasynUser->reason;
	int mode, offset;
	unsigned short sdata;

    if (this->rebooting) epicsThreadSuspendSelf();
	
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);

	this->getAddress(pasynUser, &channel);
	
    if((function == hy8414Data_) || (function == hy8414FloatData_))
    {
        pointer = this->adc->reg.pointer;
        switch(mode)
        {
            case hy8414Continuous:
                if ((size_t)pointer >= nelements) 
                    startPoint = pointer - nelements;
                else
                    startPoint = this->banksize * 2 - (nelements - pointer);
                numToReturn = nelements;
                break;
            case hy8414Triggered:
                if (pointer == 0) pointer = this->lastPointer; 	/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (pointer <= offset) 
                {
                    startPoint = 0;
                    numToReturn = pointer;
                } else
                {
                    startPoint = offset;
                    numToReturn = pointer - offset;
                }
                break;
            case hy8414Gated:
                if (pointer == 0) pointer = this->lastPointer; 	/* get pointer saved by the interrupt but only when the pointer is reset by the enable function */
                if (offset >= 0)
                {
                    if (pointer <= offset) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = offset;
                        numToReturn = pointer - offset;
                    }
                }
                else
                {
                    if (pointer < (-offset)) 
                    {
                        startPoint = 0;
                        numToReturn = pointer;
                    } else
                    {
                        startPoint = pointer + offset;
                        numToReturn = (-offset);
                    }
                }
                break;
            default:
                status = asynError;
                break;
        }
		
		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

		for(i=0; i<(int)nelements; i++)
        {
            if (i < numToReturn)
            {
            	sdata = (this->convMemoryBase + half_mem + channel * this->banksize)[startPoint];
           		*(data + i) = toVoltage((double) sdata);

/*for debug*/
//if(*(data+i) <= -9.5) printf("ch=%d i=%d p=%d sp=%d hm=%x no=%d sv=%x v=%f\n", channel, i, pointer, startPoint, half_mem, numToReturn, sdata, *(data+i));

            	startPoint++;
            }
            else
                *(data + i) = 0.0;
            if(startPoint >= this->banksize)
            {
            	startPoint = 0;                          		/* if pointer reaches the bottom of the memory, reset it */
				half_mem = half_mem == 0 ? HALF_MEM : 0;
            }
        }
        if (nelements > (size_t)numToReturn)                                            	/* if number to return less than number of elements, NORD = number of return */
            *nactual = numToReturn;
        else
            *nactual = nelements;                                               	/* if number to return is greater or equal number of elements, NORD = number of elements */
	}

	status = asynSuccess;
	return( status );
}

/*********************************************************/
/*                                                      */
/* Private Sector                                      */
/*                                                    */
/*****************************************************/

void Hy8414Asyn::setupADC( int runtime)
{
	int fastadc, enable, gain, range;
	
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8414FastADC_, &fastadc);
	getIntegerParam(hy8414Gain_, &gain);
	getIntegerParam(hy8414Range_, &range);

/* this first write to the CSR sets this bit and clears all others */
	this->adc->reg.status = 0;							 				/* 2M memory byte */
    if( fastadc )
        this->adc->reg.extStatus = 0;                                   /* normal mode doesn't need interrupt on each conversion */
    else
    	this->adc->reg.extStatus = EXTCSR_ISE;                          /* clear extension CSR and set single conversion interrupt enable for mca and EPID record support */
       
/* clear pointer to zero at start */
    this->adc->reg.pointer = 0;

/* initialise average overflow flag */
   	checkAverageOverflow( runtime );	                             	/* check average setting overflow */

/* set up the requested internal clock rate */
	setClockRate( runtime );	          

/* set up the requested external / internal clock */
	setExtClock( runtime );	              

/* set offset to check if it exceeds BANKSIZE when added to samples */
    setOffset( runtime );

/* set up samples */
	setSamples( runtime );
	
/* set ADC gain. */
	setGain( gain );

/* setup range. */
	setRange( range );

/* the mode setup of the ADC depends on the mode that has been requsted. */
	setMode( runtime );

/* disable the ADC at the beginning if it is not continuous mode and wait for the enable command */
	if(runtime)
		getIntegerParam(hy8414Enable_, &enable);
	else
		enable = ENABLEADC;
    setEnableDisable( enable );    
}

int Hy8414Asyn::setClockRate( int runtime  )
{
	int clockRate, extClock;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetClockRate_, &clockRate);
	getIntegerParam(hy8414SetExtClock_, &extClock);

	if (this->rebooting)  epicsThreadSuspendSelf();
	
	if(extClock == 1)				/* for external clock */
	{
		if(clockRate >= 15)
	        this->adc->reg.clockRate = 16;
		return(asynSuccess);
	}
	
	/* for internal clock */
    if ((clockRate < 0) || (clockRate > 16)) return( asynError );

    this->adc->reg.clockRate = clockRate;
    return(asynSuccess);
}

int Hy8414Asyn::setExtClock( int runtime )
{
	int clockRate, extClock, enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetClockRate_, &clockRate);
	getIntegerParam(hy8414SetExtClock_, &extClock);
	getIntegerParam(hy8414Enable_, &enable);

	if ( this->rebooting )  epicsThreadSuspendSelf();
       
    if( runtime )                                                   /* change samples at realtime */
    { 
        this->adc->reg.status &= ~CSR_ARM;                          /* disarm ADC first*/
        if (extClock)
	       this->adc->reg.status |= CSR_XC;
	    else
	       this->adc->reg.status &= ~CSR_XC;
        setEnableDisable(enable);          							/* start acquisition if it is enabled */
    }else
    {
        if (extClock)
	      this->adc->reg.status |= CSR_XC;
	    else
	      this->adc->reg.status &= ~CSR_XC;
    }    
    return( asynSuccess );
}

int Hy8414Asyn::softwareTrigger()
{
	int mode, enable;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414Enable_, &enable);
 
    if(mode != hy8414Triggered ) 
        return asynError;                                           /* software trigger at realtime only applies to trigger mode */
    if(enable)                                           			/* only trigger when enabled */
    {
        this->adc->reg.status &= ~CSR_ST;                           /* clear software trigger latch */
        this->adc->reg.status |= CSR_ST;                            /* software trigger to kick off the acquisition */
        this->adc->reg.status &= ~CSR_ST;                           /* clear software latch to enable hardware trigger afterwards */
		setIntegerParam(hy8414TriggerStatus_, TRIGGERRED_COLLECTING);
  		callParamCallbacks();
    }
    return( asynSuccess );
}

int Hy8414Asyn::setSamples(int runtime )
{
	int mode, average, samples, offset, enable, fastadc; 
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetSample_, &samples);
	getIntegerParam(hy8414SetAverage_, &average);
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);
	getIntegerParam(hy8414Enable_, &enable);
	getIntegerParam(hy8414FastADC_, &fastadc);

    if(runtime)                                                     /* change samples at realtime */
    { 
        switch(mode)
        {
            case hy8414Continuous:
                this->adc->reg.status &= ~CSR_ARM;                  /* disarm ADC first*/
                if(samples > (this->banksize * 2)) 
                {
                    samples = this->banksize * 2;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
                    setIntegerParam(hy8414Overflow_, 1);
                } else
                    setIntegerParam(hy8414Overflow_, 0);
                if( fastadc ) 
                	this->adc->reg.numConversions = samples;  		/* put the number of conversion in */                    
                else
                	this->sampleCounter = 0;                        /* clear sample counter */
                setEnableDisable( enable );  						/* start acquisition if it is enabled */
                break;
            case hy8414Triggered:
            	this->adc->reg.status &= ~CSR_ARM;                  /* disarm ADC first*/
                if((samples + offset) > (this->banksize * 2)) 
                {
                    samples = this->banksize * 2 - offset;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
                    setIntegerParam(hy8414Overflow_, 1);            /* samples setting exceeds limit */
                } else
                	setIntegerParam(hy8414Overflow_, 0);            /* samples setting within limit */
                this->adc->reg.numConversions = samples + offset;   /* put the new number of conversion in */

                if (average <= samples)                 			/* if average is less than samples, fine */
                	setIntegerParam(hy8414AverageOverflow_, 0);
                else                                                /* if average is greater than samples, averageoverflow */
                {
                    /*pPvt->average = pPvt->samples;*/              /* only averaging the user interested data */
                	setIntegerParam(hy8414AverageOverflow_, 1);     /* averaging overflow, errorred */
                }
                setEnableDisable( enable );  						/* enable trigger if it is allowed */
                break;
            case hy8414Gated:                                       /* gated mode has nothing to do with the samples */
                break;
            default:
                return asynError;
                break;
        }    
    } else                                                          /* normal samples setting during the start up  */
    {
       switch(mode)
       {
            case hy8414Continuous:
                if(samples > (this->banksize * 2)) 
                {
                    samples = this->banksize * 2;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
                    setIntegerParam(hy8414Overflow_, 1);            /* samples setting exceeds limit */
                } else
                	setIntegerParam(hy8414Overflow_, 0);            /* samples setting within limit */

                if( fastadc ) 
                    this->adc->reg.numConversions = samples;  		/* put the number of conversion in */                    
                else
                    this->sampleCounter = 0;                        /* clear sample counter */
                break;
            case hy8414Triggered:
                if((samples + offset) > (this->banksize * 2)) 
                {
                    samples = this->banksize * 2 - offset;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
                    setIntegerParam(hy8414Overflow_, 1);            /* samples setting exceeds limit */
                } else
                	setIntegerParam(hy8414Overflow_, 0);            /* samples setting within limit */

                if (average <= samples)                 			/* if average is less than samples, fine */
                	setIntegerParam(hy8414AverageOverflow_, 0);
                else                                                /* if average is greater than samples, averageoverflow */
                {
                    /*pPvt->average = pPvt->samples;*/              /* only averaging the user interested data */
                	setIntegerParam(hy8414AverageOverflow_, 1);     /* averaging overflow, errorred */
                }

                this->adc->reg.numConversions = samples + offset;
                break;
            case hy8414Gated:                                       /* gated mode has nothing to do with the samples */
                break;
            default:
                return asynError;
                break;
        }    
    }
	callParamCallbacks();
    return( asynSuccess );
}


int Hy8414Asyn::setMode( int runtime )
{
	int mode, enable, fastadc;
	if (this->rebooting)  epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414Enable_, &enable);
	getIntegerParam(hy8414FastADC_, &fastadc);
   
    if ((mode < hy8414Continuous) || (mode > hy8414Gated)) return(asynError);

    if( runtime ) this->adc->reg.status &= ~CSR_ARM;		        /* stop 8414 totally */

    switch(mode)
    {
        case hy8414Continuous:
            enable = ENABLEADC;                          			/* continuous mode enable adc at start */
            this->sampleCounter = 0;                                /* clear sample counter */
            if( fastadc )                                     		/* if it is normal mode, need conversion complete interrupt */
            {
                this->adc->reg.status |= CSR_EX;                    /* enable trigger */
                this->adc->reg.status |= CSR_EE;                    /* enable number of conversion interrupt */
           }
            break;
        case hy8414Triggered:                                       /* do nothing. The adc will be enabled by setEnable call */
			setIntegerParam(hy8414Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8414AutoReenable_, DISABLEADC);
            this->adc->reg.status |= CSR_EX;                        /* enable trigger */
            this->adc->reg.status |= CSR_EE;                        /* enable number of conversion interrupt */
            break;
        case hy8414Gated:
			setIntegerParam(hy8414Enable_, DISABLEADC);
			this->enable = DISABLEADC;
			setIntegerParam(hy8414AutoReenable_, DISABLEADC);
            this->adc->reg.extStatus |= EXTCSR_SMI;                 /* set software inhibit to preventing any writing to the memory */
            this->adc->reg.status |= CSR_ET;                        /* enable memory inhibit bit */
            this->adc->reg.status |= CSR_EII;                       /* enable inhibit interrupt */
            this->adc->reg.status |= CSR_ARM;	                    /* Arm the 8414 to kick off acquiring continuously */
            break;
        default:
            return asynError;
            break;
    }

	callParamCallbacks();
    return( asynSuccess );
}

int Hy8414Asyn::setRange( int range )
{
	if (this->rebooting)  epicsThreadSuspendSelf();
 
	if(range)
	    this->adc->reg.extStatus |= EXTCSR_R;                           /* clear software trigger latch */
    else
		this->adc->reg.extStatus &= ~EXTCSR_R;                            /* software trigger to kick off the acquisition */
	setIntegerParam(hy8414Range_, range);
	callParamCallbacks();
    return( asynSuccess );
}

int Hy8414Asyn::setGain( int gain )
{
	if (this->rebooting)  epicsThreadSuspendSelf();
    if ((gain < 0) || (gain > 3)) return(asynError);
 
    gain |= (gain << 2);
    this->adc->reg.gain = gain;                           /* clear software trigger latch */

    setIntegerParam(hy8414Gain_, gain);
	callParamCallbacks();
    return( asynSuccess );
}

int Hy8414Asyn::setEnableDisable( int enable )
{
	int mode, fastadc;
	if (this->rebooting)  epicsThreadSuspendSelf();

	setIntegerParam(hy8414Enable_, enable);
	this->enable = enable;
	
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414FastADC_, &fastadc);
 
	if (enable)                                                     /* if enable */
    {
        switch (mode)
        {
            case hy8414Continuous:
                this->adc->reg.status |= CSR_ARM;	                /* Arm the 8414, enable acquiring */
                if( fastadc )                                 		/* if it is normal mode, trigger the ADC */
                {
                    this->adc->reg.status &= ~CSR_ST;               /* clear software trigger */
                    this->adc->reg.status |= CSR_ST;                /* software trigger to kick off the acquisition */
                }
                break;
            case hy8414Triggered:
                this->adc->reg.status &= ~CSR_EX;                   /* disable trigger to clear any pending triggers */
                this->adc->reg.status |= CSR_EX;                    /* enable trigger */
                this->adc->reg.status |= CSR_ARM;	                /* Arm the 8414, enable acquiring */
                setIntegerParam(hy8414TriggerStatus_, WAITING_TRIGGER);
                this->adc->reg.pointer = 0;                         /* zero the address pointer */
                break;
            case hy8414Gated:
                this->adc->reg.extStatus &= ~EXTCSR_SMI;            /* clear software iiinhibit to allow hardware inhibit */
                setIntegerParam(hy8414TriggerStatus_, WAITING_TRIGGER);
                this->adc->reg.pointer = 0;                         /* zero the address pointer */
                break;
            default:
                return asynError;
                break;
        }
    }
	else                                                            /* if disable */
    {
        switch (mode)
        {
            case hy8414Continuous:                                  /* do nothing. wait for the interrupt to finish the current acquisition then stop */
            case hy8414Triggered:
                this->adc->reg.status &= ~CSR_ARM;	                /* disable 8414 */
                break;
            case hy8414Gated:
                 this->adc->reg.extStatus |= EXTCSR_SMI;            /* set software inhibit to preventing any writing to the memory */
               break;
            default:
                return asynError;
                break;
        }
    }
	
//	callParamCallbacks();
	return( asynSuccess );
}

int Hy8414Asyn::setOffset( int runtime )
{
	int mode, average, samples, offset, enable; 
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetSample_, &samples);
	getIntegerParam(hy8414SetAverage_, &average);
	getIntegerParam(hy8414SetMode_, &mode);
	getIntegerParam(hy8414SetOffset_, &offset);
	getIntegerParam(hy8414Enable_, &enable);

    if(runtime)                                                     	/* change samples at realtime */
    { 
        switch (mode)
        { 
            case hy8414Triggered:
                this->adc->reg.status &= ~CSR_ARM;	                	/* disable 8414 */
                if((samples + offset) > (this->banksize * 2)) 
                {
                	samples = this->banksize * 2 - offset;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
					setIntegerParam(hy8414Overflow_, 1);      			/* samples setting exceeds limit */
                } else
                    setIntegerParam(hy8414Overflow_, 0);            	/* samples setting within limit */
                this->adc->reg.numConversions = samples + offset;      	/* put the new number of conversion in */
                setEnableDisable(enable);  								/* enable trigger if it is allowed */
                break;
            case hy8414Gated:
                if (offset > this->banksize) 
					setIntegerParam(hy8414Overflow_, 1);      			/* samples setting exceeds limit */
                else
                    setIntegerParam(hy8414Overflow_, 0);            	/* samples setting within limit */
                break;
            default:
                return asynError;
                break;
        }
    } else
    {
        switch (mode)
        { 
            case hy8414Triggered:
                if((samples + offset) > (this->banksize * 2)) 
                {
                	samples = this->banksize * 2 - offset;
					setIntegerParam(hy8414SetSample_, samples);
					this->samples = samples;
 					setIntegerParam(hy8414Overflow_, 1);      			/* samples setting exceeds limit */
                } else
                    setIntegerParam(hy8414Overflow_, 0);            	/* samples setting within limit */
                this->adc->reg.numConversions = samples + offset;      	/* put the new number of conversion in */
                break;
            case hy8414Gated:
                if (offset > this->banksize) 
 					setIntegerParam(hy8414Overflow_, 1);      			/* samples setting exceeds limit */
                else
                   	setIntegerParam(hy8414Overflow_, 0);            	/* samples setting within limit */
                break;
            default:
                return asynError;
                break;
        }     
    }   
  	callParamCallbacks();
    return( asynSuccess );
}

int Hy8414Asyn::checkAverageOverflow( int runtime )
{
	int mode, average, samples;
    if (this->rebooting) epicsThreadSuspendSelf();

	getIntegerParam(hy8414SetSample_, &samples);
	getIntegerParam(hy8414SetAverage_, &average);
	getIntegerParam(hy8414SetMode_, &mode);

	switch(mode)
    {
        case hy8414Continuous:
            if (average > (this->banksize * 2))
            {
				setIntegerParam(hy8414AverageOverflow_, 1);
                setIntegerParam(hy8414SetAverage_, this->banksize * 2); /* average greater than the memory size, trim it to the memory size */
            }
            else
				setIntegerParam(hy8414AverageOverflow_, 0);
            break;
        case hy8414Triggered:
            if (average <= samples)                     			/* if average is less than samples, fine */
                setIntegerParam(hy8414AverageOverflow_, 0);
            else                                                    /* if average is greater than samples, averageoverflow */
            {
                /*pPvt->average = pPvt->samples;*/                      /* only averaging the user interested data */
                setIntegerParam(hy8414AverageOverflow_, 1);         /* averaging overflow, errorred */
            }
            break;
        case hy8414Gated:                                           /* gated mode has to be checked at runtime */
            if (average > (this->banksize * 2))
            {
                setIntegerParam(hy8414AverageOverflow_, 1);
                setIntegerParam(hy8414SetAverage_, this->banksize * 2); /* average greater than the memory size, trim it to the memory size */
            }
            else
                setIntegerParam(hy8414AverageOverflow_, 0);
            break;
        default:
            break;
    }    
  	callParamCallbacks();
    return( asynSuccess );
}

void Hy8414Asyn::rebootCallback()
{
    this->adc->reg.status &= ~CSR_ARM;                              	/* Clear the ARM bit  */
    this->rebooting = 1;				                            	/* flag we are rebooting */
}

/*********************************************************/
/*                                                      */
/* Floating Voltage Convertion Routines                */
/*                                                    */
/*****************************************************/

/* this routine convert the raw adc value to voltage, -10V ~ +10V or -5V ~ +5V as per the range setting.  */
double Hy8414Asyn::toVoltage( double rawval )
{
	int range;
	
	getIntegerParam(hy8414Range_, &range);

	if(range == 0)												// The range is +/-10V
	{
		return (double)10 - 20*((double)POSI_FULL_VALUE - rawval) / ((double) POSI_FULL_VALUE - NEG_FULL_VALUE);
	}else														// The range is +/-5V
	{
		return (double)5 - 10*((double)POSI_FULL_VALUE - rawval) / ((double) POSI_FULL_VALUE - NEG_FULL_VALUE);
	}    
}



/*********************************************************/
/*                                                      */
/* External Wrapped Routines      					   */
/*                                                    */
/*****************************************************/

// These functions must have C linkage because they are called from other EPICS components
extern "C" {
static void rebootCallbackC(void *drvPvt)
{
    Hy8414Asyn *pPvt = (Hy8414Asyn *)drvPvt;
    pPvt->rebootCallback();
}

static void INTSRC(int cardNumber)
{
    Hy8414Asyn *pPvt = driverTable[cardNumber];
    pPvt->intSR();
}

static void intQueuedTaskC(void *drvPvt)
{
    Hy8414Asyn *pPvt = (Hy8414Asyn *)drvPvt;
    pPvt->intQueuedTask();	
}

static void intSweepQueuedTaskC(void *drvPvt)
{
    Hy8414Asyn *pPvt = (Hy8414Asyn *)drvPvt;
    pPvt->intSweepQueuedTask();		
}

}

/*********************************************************/
/*                                                      */
/* Interrupt and Queue Task and external Routines      */
/*                                                    */
/*****************************************************/
void Hy8414Asyn::intSR()
{
    int data[2], flag = 1, sweepFlag = 1;
    int sweepData[MAX_ADC_CHANNELS + 2];
    int i = 0;
    USHORT regData;

    data[0] = this->adc->reg.pointer;                                               /* index 0 stores the address pointer */
    sweepData[0] = data[0];                                          				/* index 0 stores the address pointer */

	if (this->adc->reg.status & CSR_F)    											/* memory full flag. This interrupt doesn't do very much at the moment. Only clear flags. */
	{
		this->adc->reg.status &= ~(CSR_F | CSR_HF);									/* clear half and full flag */
//	    this->adc->reg.pointer = 0;													/* clear the pointer to force it to start from the beginning of lower half */
        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);						/* reenable interrupt */
	}
	else if (!(this->adc->reg.status & CSR_CC) && !(this->adc->reg.status & CSR_MII))    /* if it is single conversion interrupt */
    {
    	this->adc->reg.extStatus |= EXTCSR_IS;                                   	/* write 1 to clear singel conversion interrupt */
        regData = this->adc->reg.extStatus;                                         /* to just ensure write successful */
       
        switch (this->mode)
        {
            case hy8414Continuous:
            case hy8414Triggered:
	        	for (i=0; i<MAX_ADC_CHANNELS; i++)                                /* the rest of the elements store 8 channel raw data */
	        	{
    	        	sweepData[i + 2] = this->adc->reg.data[i];	        			
	        	}
                break;
            case hy8414Gated:
				if ((!(this->adc->reg.status & CSR_MIS)) 
                    && (!(this->adc->reg.extStatus & EXTCSR_SMI)))	                /* only collect data when gate is open */
				{
	        		for (i=0; i<MAX_ADC_CHANNELS; i++)                            /* the rest of the elements store 8 channel raw data */
		        	{
	    	        	sweepData[i + 2] = this->adc->reg.data[i];	        			
		        	}
				} else
					sweepFlag = 0;			      					                /* if gate not opened, don't wake up the queue */
                break;
            default:
                sweepFlag = 0;
                break;
        }

        /* Wake up sweep queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (sweepFlag)
        {
            if (epicsMessageQueueTrySend(this->intSweepQId, sweepData, sizeof(sweepData)) == 0)
            	messagesSent++;
            else
            	messagesFailed++;
        }

        if( this->mode == hy8414Continuous)                         						/* only do this when it is continuous mode and slow ADC. fastADC wouldn't get here */
        {
            sampleCounter++;
            if ( sampleCounter >= this->samples)
            {
            	sampleCounter = 0;
                /* Wake up queue task which calls callback routines */
                if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
                	messagesSent++;
                else
                	messagesFailed++;
            }
        }
    }
    else		//conversion complete or gated interrupt
    {
    	/* re-enable interrupt */
        switch (this->mode)
        {
            case hy8414Continuous:                                  				/* for mca mode (slow ADC), the code wouldn't get here */
            	this->adc->reg.status &= ~CSR_CC;                   				/* clear interrupt */
            	if(this->enable)
            	{
					this->adc->reg.status |= CSR_ARM;	                			/* Arm the 8414, enable acquiring */
                    if( this->fastadc )                                 			/* if it is normal mode, trigger the ADC */
                    {
                        this->adc->reg.status &= ~CSR_ST;               			/* clear software trigger */
                        this->adc->reg.status |= CSR_ST;                			/* software trigger to kick off the acquisition */
                    }            		
            	}else
                    this->adc->reg.status &= ~CSR_ARM;	                			/* disable 8414 */
                break;
            case hy8414Triggered:
            	this->adc->reg.status &= ~CSR_CC;                   				/* clear interrupt */
                break;
            case hy8414Gated:
            	this->adc->reg.status |= CSR_MII;                   				/* clear interrupt */
                break;
            default:
                flag = 0;
                break;
        }
        regData = this->adc->reg.status;                            				/* to just ensure write successful */

        /* Wake up queue task which calls callback routines if it is the correct interrupt. Note, the mode setting guarantees the interrupt settings */
        if (flag)
        {
            if (epicsMessageQueueTrySend(this->intMsgQId, data, sizeof(data)) == 0)
            	messagesSent++;
            else
            	messagesFailed++;
        }
    }
//    callParamCallbacks();
}

void Hy8414Asyn::intQueuedTask()
{
    int i;
    int data[2];
    epicsInt32 localValue;
    int pointer, startPoint = 0, numOfElement = 0, starttemp;
    double dValue = 0.0, fValue = 0.0;
	int mode, samples, average, offset, chan, reEnableState, half_mem = 0, half_memtemp;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intMsgQId, data, sizeof(data));
                 
    	getIntegerParam(hy8414SetMode_, &mode);
    	getIntegerParam(hy8414SetSample_, &samples);
    	getIntegerParam(hy8414SetAverage_, &average);
    	getIntegerParam(hy8414SetOffset_, &offset);
    	getIntegerParam(hy8414AutoReenable_, &reEnableState);

    	pointer = data[0];
        numOfElement = average;                                               		/* assuming average setting is what we want */
               
        switch ( mode )
        {
            case hy8414Continuous:                                                  /* continuous mode checks average overflow at runtime */
            case hy8414Triggered:                                                   /* trigger mode checks average overflow at runtime */
                if (average > samples)
                {
                    numOfElement = samples;  		                                /* use the data the user is interested */
                    setIntegerParam(hy8414AverageOverflow_, 1);
                }else
                	setIntegerParam(hy8414AverageOverflow_, 0);;                    /* normal averaging */
                if(mode == hy8414Triggered) setIntegerParam(hy8414TriggerStatus_, FINISHED_AVAILABLE);        	/* finished and data is available */
                break;
            case hy8414Gated:                                                       /* gated mode checks average overflow at runtime */
                setIntegerParam(hy8414TriggerStatus_, FINISHED_AVAILABLE);        	/* finished and data is available */
                if (offset >= 0)                                              		/* offset is possitive */
                {
                    if (pointer >= offset)                                        	/* please note the limitation of total samples of the gated mode is the BANKSIZE */
                    {
                        if (pointer - offset < average)  	                   		/* if have enough data for averaging */
                        {
                        	setIntegerParam(hy8414AverageOverflow_, 1);
                            numOfElement = pointer - offset;                  		/* use only user interested data */
                        }else
                        	setIntegerParam(hy8414AverageOverflow_, 0);             /* averaging normal */
                    }else
                    {
                    	setIntegerParam(hy8414AverageOverflow_, 1);
                        numOfElement = pointer;
                    }
                }else                                                               /* offset is negative */
                {
                    if (pointer >= (-offset))                                     	/* collected data more than offset */
                    {
                        if (average > (-offset))                           	 		/* but average is greater than offset */
                        {
                        	setIntegerParam(hy8414AverageOverflow_, 1);
                            numOfElement = (-offset);                             	/* only use data the user is interested */
                        }else
                        	setIntegerParam(hy8414AverageOverflow_, 0);;            /* averaging normal */
                    }else                                                           /* collected data less than offset */
                    {
                        numOfElement = pointer;                                     /* take everything */
                        setIntegerParam(hy8414AverageOverflow_, 1);
                    }
                }
                break;
        }

        if((data[0] - numOfElement) < 0)
           	startPoint = this->banksize * 2 - (numOfElement - data[0]);
        else
        	startPoint = data[0] - numOfElement;
        	
		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									/* if pointer reaches the end of the either the full memory of half full, reset it */
			half_mem = HALF_MEM;
        }

		half_memtemp = half_mem;
        starttemp = startPoint;
  
        for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
        	/* Pass int32 interrupts, Pass float64 interrupts */
            dValue = 0.0;
            fValue = 0.0;
            startPoint = starttemp;
			half_mem = half_memtemp;
            for(i=0; i<numOfElement; i++)
            {
                dValue += (double) (this->convMemoryBase + half_mem + chan * this->banksize)[startPoint];
                startPoint++;            		
                if(startPoint >= this->banksize)
                {
                	startPoint = 0;                          		/* if pointer reaches the bottom of the memory, reset it */
    				half_mem = half_mem == 0 ? HALF_MEM : 0;
                }
            }

            if (numOfElement != 0)                                              	/* safety check */
                localValue = (epicsInt32) (dValue / numOfElement);
            else
                localValue = (epicsInt32) dValue; 
    		fValue = toVoltage(localValue);
           
    		setIntegerParam(chan, hy8414Data_, localValue);
    		setDoubleParam(chan, hy8414FloatData_, fValue);

			callParamCallbacks(chan);

        }

       /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/
        pointer = data[0];
        this->lastPointer = data[0];                                                		/* store pointer for periodic records (poll mode) */

        switch(mode)
        {
            case hy8414Continuous:
                if (pointer >= samples) 
                    startPoint = pointer - samples;
                else
                    startPoint = this->banksize * 2 - (samples - pointer);
                numOfElement = samples;
                break;
            case hy8414Triggered:
                startPoint = pointer - samples;
                numOfElement = samples;
                break;
            case hy8414Gated:
                if (offset >= 0)                                              		/* offset is positive */
                {
                    if (pointer > offset)                                     		/* number of data collected is greater than offset */
                    {
                        startPoint = offset;
                        numOfElement = pointer - offset;
                    }else                                                           /* number of data collected is less than offset, no enough data, return all collected */
                    {
                        startPoint = 0;
                        numOfElement = pointer;
                    }
                }else                                                               /* offset is negative */
                {
                    if (pointer > (-offset))
                    {
                        startPoint = pointer + offset;
                        numOfElement = -offset;                 
                    }else
                    {
                        startPoint = 0;
                        numOfElement = pointer;                 
                    }
                }
                break;
        }

		half_mem = 0;
        if(startPoint >= this->banksize)
        {
        	startPoint = startPoint - this->banksize;      									// if pointer reaches the end of the either the full memory of half full, reset it
			half_mem = HALF_MEM;
        }

/*for debug*/
//printf("csr=%x ecsr=%x p=%d sp=%d hm=%x \n", this->adc->reg.status, this->adc->reg.extStatus, pointer, startPoint, half_mem);
     
		half_memtemp = half_mem;
        starttemp = startPoint;
        for(chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
        	// Pass int32Array interrupts 
            startPoint = starttemp;
			half_mem = half_memtemp;
            for(i=0; i<numOfElement; i++)
            {
           		rawData[chan][i] = (ULONG) (this->convMemoryBase + chan * this->banksize)[startPoint];
            	adcData[chan][i] = toVoltage((double)rawData[chan][i]);
            	startPoint++;
                if(startPoint >= this->banksize)
                {
                	startPoint = 0;                          							// if pointer reaches the bottom of the memory, reset it
    				half_mem = half_mem == 0 ? HALF_MEM : 0;
                }
            }

			doCallbacksInt32Array((epicsInt32 *) &rawData[chan],
                                        numOfElement, hy8414Data_, chan);
 			doCallbacksFloat64Array((epicsFloat64 *) &adcData[chan],
                                        numOfElement, hy8414FloatData_, chan);
        }
        /* re-enable interrupt */
        switch (mode)
        {
            case hy8414Continuous:                                                  /* continuous mode re-initialise the adc in isr. Here nothing needs to be done */
                break;
            case hy8414Triggered:
            case hy8414Gated:
                if (!reEnableState)                                           		/* if re-trigger is disabled, stop further triggers. To trigger again, need to do an enable command */
                    setEnableDisable( DISABLEADC );                       			/* disable trigger */
                else
                    setEnableDisable( ENABLEADC );                        			/* enable trigger */
                break;
            default:
                break;
        }
        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);
	}
    callParamCallbacks();
}


void Hy8414Asyn::intSweepQueuedTask()
{
    int sweepData[MAX_ADC_CHANNELS + 2];
    int chan;
    double dValue = 0.0;

    while(1) {
        /* Wait for event from interrupt routine */
        epicsMessageQueueReceive(intSweepQId, sweepData, sizeof(sweepData));

        /* Pass float64 interrupts */
        for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
            dValue = (double) sweepData[chan + 2];
    		setDoubleParam(chan, hy8414FloatSweepData_, dValue);
			callParamCallbacks(chan);
        }

        /****************************************/
        /****************************************/
        /********* waveform collection **********/
        /****************************************/
        /****************************************/

        /* Pass int32Array interrupts */
        for (chan=0; chan<MAX_ADC_CHANNELS; chan++)
        {
            chanData[chan] = (ULONG) sweepData[chan + 2];
        }
		doCallbacksInt32Array((epicsInt32 *) &chanData,
				MAX_ADC_CHANNELS, hy8414SweepData_, 0);

        ipmIrqCmd(carrierNum, ipSlotNum, 0, ipac_irqEnable);

    }
}

